/*
 * Copyright (c) 2010-2012 The Amdatu Foundation
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.amdatu.web.resourcehandler.itest;

import java.io.IOException;
import java.net.HttpURLConnection;
import java.net.URL;
import java.util.concurrent.CountDownLatch;
import java.util.concurrent.TimeUnit;

import junit.framework.TestCase;

import org.osgi.framework.Bundle;
import org.osgi.framework.BundleContext;
import org.osgi.framework.BundleEvent;
import org.osgi.framework.BundleListener;
import org.osgi.framework.FrameworkUtil;

public class ResourceHandlerTest extends TestCase implements BundleListener {
	/**
	 * Provides a custom {@link CountDownLatch} that allows it to count down in 
	 * case a bundle event for a specified bundle with a specified bundle 
	 * event-type is given.
	 */
	static class BundleStateLatch {
		private final CountDownLatch m_latch;
		private final Bundle m_bundle;
		private final int m_type;

		public BundleStateLatch(Bundle bundle, int type) {
			m_latch = new CountDownLatch(1);
			m_bundle = bundle;
			m_type = type;
		}
		
		/**
		 * Awaits until this latch reaches a count of zero or when 5 seconds 
		 * are passed, whichever comes first.
		 */
		public boolean await() throws InterruptedException {
			return m_latch.await(5, TimeUnit.SECONDS);
		}
		
		/**
		 * Counts down in case the given bundle event matches the contained 
		 * bundle and type.
		 * 
		 * @param event the bundle event to test, cannot be <code>null</code>.
		 * @return <code>true</code> if the event matched, <code>false</code> otherwise.
		 */
		public boolean countDownOnMatch(BundleEvent event) {
			if (event.getBundle() == m_bundle && event.getType() == m_type) {
				m_latch.countDown();
				return true;
			}
			return false;
		}
	}
	
	private static final int HTTP_OK = 200;
	private static final int HTTP_NOT_FOUND = 404;
	
	private static final String URL_HELLO_WORLD = "http://localhost:8080/bundle1/helloworld.txt";
	private static final String URL_FAREWELL_WORLD = "http://localhost:8080/bundle2/farewellworld.txt";
	private static final String URL_FAREWELL_AGAIN = "http://localhost:8080/bundle2again/farewellworld.txt";
	private static final String URL_FAREWELL_HELLO_WORLD = "http://localhost:8080/bundle1/bundle2/farewellworld.txt";
	private static final String URL_WOOT = "http://localhost:8080/bundle3/woot.txt";

	private Bundle m_helloBundle;
	private Bundle m_farewellBundle;
	private Bundle m_resourceHandler;
	private Bundle m_bundleOwnContext;

	private final BundleContext m_context;

	private volatile BundleStateLatch m_latch;

	public ResourceHandlerTest() {
		m_context = FrameworkUtil.getBundle(getClass()).getBundleContext();
	}

	public void bundleChanged(BundleEvent event) {
		if (m_latch != null) {
			m_latch.countDownOnMatch(event);
		}
	}

	public void testHelloWorld() throws Exception {
		assertEquals(HTTP_OK, getResponseCode(URL_HELLO_WORLD));
	}

	public void testMultiUrls() throws Exception {
		assertEquals(HTTP_OK, getResponseCode(URL_FAREWELL_WORLD));
		assertEquals(HTTP_OK, getResponseCode(URL_FAREWELL_AGAIN));
		assertEquals(HTTP_OK, getResponseCode(URL_FAREWELL_HELLO_WORLD));
	}

	public void testOwnContext() throws Exception {
		assertEquals(HTTP_OK, getResponseCode(URL_WOOT));
	}

	public void testRestartResource() throws Exception {
		assertEquals(HTTP_OK, getResponseCode(URL_HELLO_WORLD));
		assertEquals(HTTP_OK, getResponseCode(URL_WOOT));
		assertEquals(HTTP_OK, getResponseCode(URL_FAREWELL_WORLD));

		stopBundle(m_helloBundle);

		assertEquals(HTTP_NOT_FOUND, getResponseCode(URL_HELLO_WORLD));
		assertEquals(HTTP_OK, getResponseCode(URL_WOOT));
		assertEquals(HTTP_OK, getResponseCode(URL_FAREWELL_WORLD));

		startBundle(m_helloBundle);

		assertEquals(HTTP_OK, getResponseCode(URL_HELLO_WORLD));
		assertEquals(HTTP_OK, getResponseCode(URL_WOOT));
		assertEquals(HTTP_OK, getResponseCode(URL_FAREWELL_WORLD));
	}

	public void testRestartResourceHandler() throws Exception {
		assertEquals(HTTP_OK, getResponseCode(URL_WOOT));
		assertEquals(HTTP_OK, getResponseCode(URL_HELLO_WORLD));
		assertEquals(HTTP_OK, getResponseCode(URL_FAREWELL_WORLD));
		
		stopBundle(m_resourceHandler);
		
		assertEquals(HTTP_NOT_FOUND, getResponseCode(URL_WOOT));
		assertEquals(HTTP_NOT_FOUND, getResponseCode(URL_HELLO_WORLD));
		assertEquals(HTTP_NOT_FOUND, getResponseCode(URL_FAREWELL_WORLD));

		startBundle(m_resourceHandler);
		
		assertEquals(HTTP_OK, getResponseCode(URL_WOOT));
		assertEquals(HTTP_OK, getResponseCode(URL_HELLO_WORLD));
		assertEquals(HTTP_OK, getResponseCode(URL_FAREWELL_WORLD));
	}

	public void testStopOwnContext() throws Exception {
		assertEquals(HTTP_OK, getResponseCode(URL_WOOT));
		assertEquals(HTTP_OK, getResponseCode(URL_HELLO_WORLD));
		assertEquals(HTTP_OK, getResponseCode(URL_FAREWELL_WORLD));

		stopBundle(m_bundleOwnContext);

		assertEquals(HTTP_NOT_FOUND, getResponseCode(URL_WOOT));
		assertEquals(HTTP_OK, getResponseCode(URL_HELLO_WORLD));
		assertEquals(HTTP_OK, getResponseCode(URL_FAREWELL_WORLD));

		startBundle(m_bundleOwnContext);

		assertEquals(HTTP_OK, getResponseCode(URL_WOOT));
		assertEquals(HTTP_OK, getResponseCode(URL_HELLO_WORLD));
		assertEquals(HTTP_OK, getResponseCode(URL_FAREWELL_WORLD));
	}

	/**
	 * Set up for each individual test.
	 */
	@Override
	protected void setUp() throws Exception {
		Bundle[] bundles = m_context.getBundles();
		for (Bundle bundle : bundles) {
			if ("org.amdatu.web.resourcehandler".equals(bundle.getSymbolicName())) {
				m_resourceHandler = bundle;
			}
			if ("org.amdatu.web.resourcehandler.itest.bundle1".equals(bundle.getSymbolicName())) {
				m_helloBundle = bundle;
			}
			if ("org.amdatu.web.resourcehandler.itest.bundle2".equals(bundle.getSymbolicName())) {
				m_farewellBundle = bundle;
			}
			if ("org.amdatu.web.resourcehandler.itest.bundle3".equals(bundle.getSymbolicName())) {
				m_bundleOwnContext = bundle;
			}
		}
		
		// Sanity check before entering any test...
		assertNotNull("...itest.bundle1 does not appear to be installed?!", m_helloBundle);
		assertNotNull("...itest.bundle2 does not appear to be installed?!", m_farewellBundle);
		assertNotNull("...itest.bundle3 does not appear to be installed?!", m_bundleOwnContext);
		assertNotNull("...web.resourcehandler does not appear to be installed?!", m_resourceHandler);

		// Wait until the resources are all registered...
		waitUntilResourcesAreAvailable();
		
		// Listen to all bundle events...
		m_context.addBundleListener(this);
	}
	
	/**
	 * Tear down for each individual test.
	 */
	@Override
	protected void tearDown() throws Exception {
		// Stop listening to bundle events...
		m_context.removeBundleListener(this);
		// Make sure we don't reuse old latches...
		m_latch = null;
	}

	/**
	 * Starts the given bundle and waits until the framework notifies all
	 * listeners that it is started.
	 */
	private void startBundle(Bundle bundle) throws Exception {
		m_latch = new BundleStateLatch(bundle, BundleEvent.STARTED);

		bundle.start();

		assertTrue("Failed to start bundle " + bundle, m_latch.await());
	}

	/**
	 * Stops the given bundle and waits until the framework notifies all
	 * listeners that it is stopped.
	 */
	private void stopBundle(Bundle bundle) throws Exception {
		m_latch = new BundleStateLatch(bundle, BundleEvent.STOPPED);
		bundle.stop();
		assertTrue("Failed to stop bundle " + bundle, m_latch.await());
	}

	/**
	 * Checks whether all desired resources are present at start up.
	 */
	private void waitUntilResourcesAreAvailable() throws Exception {
		for (int i = 0; i < 10; i++) {
			try {
				if ((getResponseCode(URL_HELLO_WORLD) == HTTP_OK)
						&& (getResponseCode(URL_WOOT) == HTTP_OK)
						&& (getResponseCode(URL_FAREWELL_WORLD) == HTTP_OK)
						&& (getResponseCode(URL_FAREWELL_HELLO_WORLD) == HTTP_OK)
						&& (getResponseCode(URL_FAREWELL_AGAIN) == HTTP_OK)) {
					return;
				}
			} catch (IOException ex) {
				// Ignore for now...
			}
			Thread.sleep(100);
		}
		fail();
	}

	/**
	 * @return the HTTP response code for the given URL, e.g., 200 or 404.
	 */
	private int getResponseCode(String urlToRead) throws Exception {
		URL url = new URL(urlToRead);
		HttpURLConnection conn = (HttpURLConnection) url.openConnection();
		conn.setRequestMethod("GET");
		return conn.getResponseCode();
	}
}
