/*
 * Copyright (c) 2010-2012 The Amdatu Foundation
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.amdatu.web.resourcehandler;

import static org.amdatu.web.resourcehandler.Constants.CONTEXTID;

import java.util.List;
import java.util.Map;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.ConcurrentMap;

import org.amdatu.web.resourcehandler.webresourcekeyparser.InvalidEntryException;
import org.amdatu.web.resourcehandler.webresourcekeyparser.RegisterItem;
import org.amdatu.web.resourcehandler.webresourcekeyparser.WebResourceKeyParser;
import org.apache.felix.dm.Component;
import org.apache.felix.dm.DependencyManager;
import org.osgi.framework.Bundle;
import org.osgi.service.http.HttpContext;
import org.osgi.service.http.HttpService;
import org.osgi.service.log.LogService;

public class ResourceHandler {

    private final ConcurrentMap<Bundle, DependencyManager> m_components;
    private final WebResourceKeyParser m_parser;

    public ResourceHandler() {
        m_components = new ConcurrentHashMap<Bundle, DependencyManager>();
        m_parser = new WebResourceKeyParser();
    }

    /**
	 * Adds the resources of the given bundle to a specified or the default
	 * HttpContext.
	 * 
	 * @param bundle
	 *            the bundle to add the resources for, cannot be
	 *            <code>null</code>.
	 * @throws InvalidEntryException
	 *             in case the bundle specified an invalid web resource key.
	 */
    public void addResourceBundle(Bundle bundle) throws InvalidEntryException {
        Map<String, List<RegisterItem>> entries = m_parser.getEntries(bundle);
        
        for (Map.Entry<String, List<RegisterItem>> entryItem : entries.entrySet()) {
        	String contextId = entryItem.getKey();
        	List<RegisterItem> registerItems = entryItem.getValue();
        	
            DependencyManager manager = new DependencyManager(bundle.getBundleContext());

            Component component = manager.createComponent()
                .setImplementation(new ResourceRegister(registerItems))
                .add(manager.createServiceDependency()
                		.setService(HttpService.class)
                		.setRequired(true))
                .add(manager.createServiceDependency()
                		.setService(LogService.class)
                		.setRequired(false));

            if ((contextId != null) && !"".equals(contextId)) {
                component.add(manager.createServiceDependency()
                    .setService(HttpContext.class, "(" + CONTEXTID + "=" + contextId + ")")
                    .setRequired(true));
            }
            manager.add(component);

            m_components.putIfAbsent(bundle, manager);
        }
    }

    /**
	 * Removes a given bundle as resource provider to the HTTP context.
	 * 
	 * @param bundle
	 *            the bundle to remove as resource provider, cannot be
	 *            <code>null</code>.
	 */
    public void removeResourceBundle(Bundle bundle) {
        DependencyManager dependencyManager = m_components.remove(bundle);
        if (dependencyManager != null) {
            dependencyManager.clear();
        }
    }
}
