/*
 * Copyright (c) 2010-2012 The Amdatu Foundation
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.amdatu.web.resourcehandler.webresourcekeyparser;

import static org.amdatu.web.resourcehandler.Constants.WEB_RESOURCE_KEY;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Scanner;

import org.osgi.framework.Bundle;

/**
 * Used to parse the WebResourceKey in a bundle manifest
 */
public class WebResourceKeyParser {

	/**
	 * Parses the entries for WebResourceKey in the given bundle.
	 * 
	 * @param bundle
	 *            the bundle to parse the WebResourceKey entries for, cannot be <code>null</code>.
	 * @return a map of entries, the key denotes the context ID and the value is a list of register items.
	 * @throws InvalidEntryException
	 *             when a entry in the key is invalid
	 */
	public Map<String, List<RegisterItem>> getEntries(Bundle bundle) throws InvalidEntryException {
		return getEntries((String) bundle.getHeaders().get(WEB_RESOURCE_KEY));
	}

	/**
	 * Parses the entries for the given WebResourceKey value.
	 * 
	 * @param key
	 *            the string value of the X-Web-Resource
	 * @return a map of entries, the key denotes the context ID and the value is a list of register items.
	 * @throws InvalidEntryException
	 *             when a entry in the key is invalid
	 */
	public Map<String, List<RegisterItem>> getEntries(String key) throws InvalidEntryException {
		Map<String, List<RegisterItem>> entries = new HashMap<String, List<RegisterItem>>();

		if ((key != null) && !"".equals(key.trim())) {
			Scanner scanner = new Scanner(key);
			scanner.useDelimiter(",\\s*");

			while (scanner.hasNext()) {
				String entry = scanner.next();

				RegisterItem item = createRegisterItem(entry);

				List<RegisterItem> items = entries.get(item.getContextId());
				if (items == null) {
					items = new ArrayList<RegisterItem>();
					entries.put(item.getContextId(), items);
				}
				items.add(item);
			}

			scanner.close();
		}

		return entries;
	}

	/**
	 * Creates a new RegisterItem with the information of the entry
	 * 
	 * @param entry
	 *            can be a combination of alias, alias;path or
	 *            alias;path;contextId
	 * @return RegisterItem with the information of the entry
	 * @throws InvalidEntryException
	 *             when the entry isn't correctly formatted
	 */
	private RegisterItem createRegisterItem(String entry) throws InvalidEntryException {
		String[] values = entry.split("\\s*;\\s*");
		if (values.length == 0 || values.length > 3) {
			throw new InvalidEntryException(entry);
		}
		for (String value : values) {
			if ("".equals(value.trim())) {
				throw new InvalidEntryException(entry);
			}
		}

		RegisterItem item = null;
		if (values.length == 1) { 
			// alias
			item = new RegisterItem(values[0]);
		} else if (values.length == 2) { 
			// alias;path
			item = new RegisterItem(values[0], values[1]);
		} else if (values.length == 3) { 
			// alias;path;contexID
			item = new RegisterItem(values[0], values[1], values[2]);
		}
		return item;
	}
}
